# vim:ts=4

#*********************************************************************
#*    PGMID.        ISO 7816-3 T=0 SEND/RECEIVE FUNCTIONS.           *
#*    AUTHOR.       BERND R. FIX   >Y<                               *
#*    DATE WRITTEN. 07/02/10.                                        *
#*    COPYRIGHT.    (SEE BELOW).                                     *
#*********************************************************************

# Simple Operating system for Smart cards:
#    Copyright (C) 2002  Matthias Bruestle <m@mbsks.franken.de>
# Modifications and bugfixes for AVR utility libraray:
#    Copyright (C) 2007  Bernd R. Fix <Bernd.Fix@aspector.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#
# $Log: T0.S,v $
# Revision 1.3  2007/02/17 16:52:11  y67
#

			.arch		avr5
			.include	"m163def.inc"

#======================================================================
#	Register assignments
#======================================================================

.equ	ZERO,		25		; local constant 0
.equ	PARAM,		24		; argument to / result of - function
.equ 	DELAY,		23		; delay parameter
.equ	PARITY,		22		; parity
.equ	COUNT,		21		; counter
.equ	TEMP,		20		; temporary

#======================================================================
#	I/O port pin definition
#======================================================================

.equ	IO_PIN,		6		; PIN 6 is used for bidirectional I/O

#======================================================================
# Serial communication (half-duplex, 9600 Bd, 8 bits, even parity,
# one stop bit) with T=0 protocol as specified in ISO 7816-3:
#
#  (b0,..,b7) = bits of transfered byte
#          bp = parity bit (even)
#		   bs = stop bit 
#          be = error indicator (protocol T0)
#
#         Start                          Parity             Next
#          bit <----- 8 data bits -----> bit              Start bit
#  Z   ____     ___________________________________......______     __
#          |   |  |  |  |  |  |  |  |  |  |  |  |              |   |
#  I/O     |   |b0|b1|b2|b3|b4|b5|b6|b7|bp|bs|be|   Guardtime  |   |
#          |___|__|__|__|__|__|__|__|__|__|  |__|              |___|_
#  A       :   :                    :     :  :                 :
#          t0  t1                   :    t10 t11               :
#          :                        :        :                 :
#          :<---- (n+/-0.2) etu --->:        :                 :
#          :                                 :                 :
#          :<-----(11+/-0.2) etu ----------->:                 :
#          :<---------------------min 12*etu ----------------->:
#
#======================================================================

			.text
			.global initT0
			.global receiveT0, sendT0
			.global ireceiveT0, isendT0
			.comm	mode, 1, 1

#======================================================================
# Timed delays -- total cycle cost of call is  (3*n + 10) 
#   [IN]  wait parameter n in register DELAY
# Remember: We have an external CLK running at 3.58 MHz
#======================================================================

#										cycles
delay:		dec		DELAY			;	1
			brne	delay			;	1/!2
			ret						;	5

#----------------------------------------------------------------------
# Delay macro: delay for given number of s
# Minimum wait time is 4 s, maximum is 216 s
#----------------------------------------------------------------------
			
.macro		Wait	t
			ldi		DELAY, lo8 ((358*\t - 1000)/300)
			rcall	delay
.endm

#======================================================================
# Convert R24 between direct and indirect convention
#======================================================================

iconvert:	push	TEMP			; save spoiled registers
			push	COUNT
			
			ldi		COUNT,8			; process eight bits
1:			ror		PARAM			; inverse bit direction...
			rol		TEMP
			dec		COUNT
			brne	1b
			
			ser		PARAM
			eor		PARAM, TEMP
			pop 	COUNT
			pop		TEMP
			ret

#######################################################################
# Initialize T=0 serial communication
#######################################################################

initT0:		push	TEMP

			ldi		TEMP, (1<<ACD)		; switch off A/D
			out		ACSR, TEMP
			
			ser		TEMP				; set port
    		out		DDRA, TEMP
    		out		DDRB, TEMP
    		out		DDRC, TEMP
    		out		DDRD, TEMP
    		out		PORTA,TEMP
    		out		PORTB,TEMP
    		out		PORTC,TEMP
    		out		PORTD,TEMP
    		pop		TEMP
    		ret

#######################################################################
# Receive a byte with T=0 error correction --> [R24:R25]
# Returns data interpreted as indirect convention!
#######################################################################

ireceiveT0:	push	PARITY
			ldi		PARITY,1
			rcall	receive
			pop		PARITY
			rjmp	iconvert

#######################################################################
# Receive a byte with T=0 error correction --> [R24:R25]
# Returns data interpreted as direct convention!
#######################################################################

receiveT0:	push	PARITY
			clr		PARITY
			rcall	receive
			pop		PARITY
			ret
#----------------------------------------------------------------------
# low-level receive
#----------------------------------------------------------------------

receive:	push	ZERO			; save spoiled registers
			push	DELAY
			push	COUNT
			push	TEMP
			push	PARITY

			clr		ZERO			; set constant 0

#----------------------------------------------------------------------
# start receive process: prepare result and port
#----------------------------------------------------------------------

rec_start:	clr		PARAM			; clear result
			
			ldi		TEMP, 1			; we are in receive mode
			sts		mode, TEMP
			
			cbi		DDRB, IO_PIN	; set pin for input
			cbi		PORTB, IO_PIN	; force state	

1:			sbic	PINB, IO_PIN	; wait for begin of start bit
			 rjmp	1b				; (skip jitter)
			sbic	PINB, IO_PIN
			 rjmp	1b

#----------------------------------------------------------------------
# start of character transmission
#
#      ____      _________________________________________________
#          |    |    |    |    |    |    |    |    |    |    |
#  I/O     |    |    |    |    |    |    |    |    |    |    |
#          |____|____|____|____|____|____|____|____|____|____|
#          ::  : :  : :  : :  : :  : :  : :  : :  : :  : :  :
#          AB  C D  E D  E D  E ...
#----------------------------------------------------------------------

			Wait	26				; [A] wait etu/4
			rcall	getbit			; [B] get start bit (returned in carry)
			brcs	1b				; [C] invalid start bit --> restart
	
			ldi		COUNT, 8		;     receive next eight bits
2:			Wait	54				;     wait etu/2
			rcall	getbit			; [D] get bit in carry flag (costs etu/2)
			ror		PARAM			; [E] shift into result
			sbrc	PARAM,7			;     if inserted bit is set...
			 inc	PARITY			;      ... increment parity counter
			
			dec		COUNT			; receive next bit
			brne	2b

#----------------------------------------------------------------------
# process parity bit
#----------------------------------------------------------------------

			Wait	54				;     wait etu/2
			rcall	getbit			; [D] get parity bit in carry flag (costs etu/2)
	
			adc		PARITY, ZERO
			sbrc	PARITY, 0		; if count (incl. parity bit and preset) is odd...
			 rjmp	rec_error		; ... we have an error (wrong parity)

#----------------------------------------------------------------------
# byte received successfully: spent guard time 
#----------------------------------------------------------------------
			Wait	66

			pop		PARITY
			pop		TEMP
			pop		COUNT
			pop		DELAY
			pop		ZERO
			ret

#----------------------------------------------------------------------
# receive error: signal error to sender as defined for protocol T0
#----------------------------------------------------------------------

rec_error:	Wait	66				; wait etu/2
			sbi		DDRB, IO_PIN	; set for error output
			cbi		PORTB, IO_PIN	; signal error
			Wait	155				; wait 1.5*etu
			pop		PARITY			; restore parity preset from stack
			push	PARITY
			rjmp	rec_start		; retry... 

#----------------------------------------------------------------------
# Get bit state into carry (takes etu/2): Take three samples
# seperated by etu/4, so overall cost is etu/2
#----------------------------------------------------------------------

getbit:		clr		TEMP			; clear sample

			sbic	PINB, IO_PIN
			 inc	TEMP
			Wait	25
			sbic	PINB, IO_PIN
			 inc	TEMP
			Wait	25
			sbic	PINB, IO_PIN
			 inc	TEMP
			
			lsr		TEMP			; majority wins -- return bit 2^1
			lsr		TEMP
			ret
	
#######################################################################
# Send a byte with T=0 error correction  <-- [R24]
# The byte sent follows indirect convention!
#######################################################################

isendT0:	push	PARITY
			ldi		PARITY,1
			rcall	iconvert		; convert to indirect convention
			rcall	send			; send out
			pop		PARITY
			ret

#######################################################################
# Send a byte with T=0 error correction  <-- [R24]
# The byte sent follows direct convention!
#######################################################################

sendT0:		push	PARITY
			clr		PARITY
			rcall	send			; send out
			pop		PARITY
			ret
			
#----------------------------------------------------------------------
# low-level send routine
#----------------------------------------------------------------------

send:		push	COUNT
			push	TEMP
			push	PARITY
			
			lds		TEMP, mode		; already in send mode?
			tst		TEMP
			breq	send_start
			
			Wait	208				; no -- wait 2*etu
			clr		TEMP			; indicate send mode
			sts		mode, TEMP

send_start:	push	PARAM			; save byte for resent
	
			sbi		PORTB, IO_PIN	; force state
			sbi		DDRB, IO_PIN	; set pin for output
	
			cbi		PORTB, IO_PIN	; Send start bit
			Wait	102				; wait etu
			
			ldi		COUNT, 8		; Send 8 bits
1:			ror		PARAM			; shift senf bit into carry
			brcs	2f				; bit clear?
			cbi		PORTB, IO_PIN	; yes -- state 0 on port
			rjmp	3f
2:			sbi		PORTB, IO_PIN   ; no -- state 1 on port
			 inc	PARITY			; increment set bit count
3:			Wait	102				; wait etu
			
			dec		COUNT			; more bits to send?
			brne	1b

			sbrc	PARITY, 0		; send parity bit
			 rjmp	4f
			cbi		PORTB, IO_PIN	; send 0 on even parity
			rjmp	5f
4:			sbi		PORTB, IO_PIN	; send 1 on odd parity
5:			Wait	102				; wait etu

			sbi		PORTB, IO_PIN	; send stop bit
			Wait	102				; wait etu
			
#----------------------------------------------------------------------
# Check for error signal from receiver (protocol T0)
#----------------------------------------------------------------------

			cbi		DDRB, IO_PIN	; switch to input
			cbi		PORTB, IO_PIN	; force state

			clc						; detect signal
			sbic	PINB, IO_PIN
			sec
			brcs	send_done		; no error

			pop		PARAM			; get saved byte value
			pop		PARITY			; restore parity preset from stack
			push	PARITY
6:			sbic	PINB, IO_PIN	; Wait for end of error signal
			rjmp	6b
			Wait	57				; wait etu/2
			rjmp	send_start		; start resending byte

send_done:	Wait	116
			pop		TEMP			; skip saved byte
			pop		PARITY
			pop		TEMP
			pop		COUNT
			ret
